/*
**  LocalFolder+maildir.m
**
**  Copyright (c) 2004
**
**  Author: Ujwal S. Sathyam <ujwal@setlurgroup.com>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**  
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**  
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <Pantomime/LocalFolder+maildir.h>

#include <Pantomime/Flags.h>
#include <Pantomime/LocalCacheManager.h>
#include <Pantomime/LocalFolder+mbox.h>
#include <Pantomime/LocalMessage.h>
#include <Pantomime/LocalStore.h>
#include <Pantomime/NSString+Extensions.h>

#include <Foundation/NSFileManager.h>
#include <Foundation/NSNotification.h>

#warning FIXME
#undef PERFORM_SELECTOR_2

#define PERFORM_SELECTOR_2(sel, name, obj, key) \
if ([(LocalStore *)[self store] delegate] && [[(LocalStore *)[self store] delegate] respondsToSelector: sel]) \
{ \
  [[(LocalStore *)[self store] delegate] performSelector: sel \
	     withObject: [NSNotification notificationWithName: name \
					 object: [self store] \
					 userInfo: [NSDictionary dictionaryWithObject: obj \
								 forKey: key]]]; \
}


//
//
//
@implementation LocalFolder (maildir)

- (void) expunge_maildir: (BOOL) returnDeletedMessages
{
  NSMutableArray *aMutableArray;
  LocalMessage *aMessage;
  Flags *theFlags;
  int i, messageNumber;
  
  aMutableArray = [NSMutableArray array];
  
  // We assume that our write operation was successful and we initialize our messageNumber to 1
  messageNumber = 1;
  
  for (i = 0; i < [allMessages count]; i++)
    {
      aMessage = [allMessages objectAtIndex: i];
      
      theFlags = [aMessage flags];
      
      if ( [theFlags contain: PantomimeDeleted] )
	{
	  // We add our message to our array of deleted messages, if we need to.
	  if ( returnDeletedMessages )
	    {
	      [aMutableArray addObject: [aMessage rawSource]];
	    }
	  
	  // Delete the message file
	  [[NSFileManager defaultManager] removeFileAtPath: [aMessage mailFilename]  handler: nil];
	  
	  // Remove the message from the cache
	  [(LocalCacheManager *)_cacheManager removeObject: aMessage];			
	}
      else
	{
	  // rewrite the message to account for changes in the flags
	  BOOL moveWasSuccessful;
	  NSString *uniquePattern, *newFileName;
	  NSMutableString *info;
	  int indexOfPatternSeparator;
	  Flags *theFlags;
	  

	  // We update our message's ivars (folder and size don't change)
	  [aMessage setMessageNumber: messageNumber];

	  // We increment our messageNumber local variable
	  messageNumber++;

	  // we rename the message according to the maildir spec by appending the status information the name
	  // name of file will be unique_pattern:info with the status flags in the info field
	  indexOfPatternSeparator = [[aMessage mailFilename] indexOfCharacter: ':'];
	  
	  if (indexOfPatternSeparator > 1)
	    {
	      uniquePattern = [[aMessage mailFilename] substringToIndex: indexOfPatternSeparator];
	    }
	  else
	    {
	      uniquePattern = [aMessage mailFilename];
	    }

	  // build the info field
	  info = [[NSMutableString alloc] initWithString: @"2,"];
	  
	  theFlags = [aMessage flags];
	  
	  if ([theFlags contain: PantomimeDraft])
	    {
	      [info appendString: @"D"];
	    }
	  
	  if ([theFlags contain: PantomimeFlagged])
	    {
	      [info appendString: @"F"];
	    }
	  
	  if ([theFlags contain: PantomimeAnswered])
	    {
	      [info appendString: @"R"];
	    }
	  
	  if ([theFlags contain: PantomimeSeen])
	    {
	      [info appendString: @"S"];
	    }
	  
#warning useless?
	  if ([theFlags contain: PantomimeDeleted])
	    {
	      [info appendString: @"T"];
	    }
	  
	  // build the new file name
	  newFileName = [NSString stringWithFormat: @"%@:%@", uniquePattern, info];
	  RELEASE(info);

	  // rename the message file
	  moveWasSuccessful = [[NSFileManager defaultManager] movePath: [aMessage mailFilename] toPath: newFileName handler: nil];
	  
	  if (moveWasSuccessful)
	    {
	      [aMessage setMailFilename: newFileName];
	    }
	}
    }
    
    // We sync our cache
    [_cacheManager synchronize];
    [self setMessages: [_cacheManager cache]];

    PERFORM_SELECTOR_2(@selector(folderExpungeCompleted:), @"PantomimeFolderExpungeCompleted", self, @"Folder");
}


//
// This parses a local structure for messages by looking in the "cur" and "new" sub-directories.
//
- (BOOL) parse_maildir: (NSString *) theDirectory
{
  NSMutableArray *mailFiles;
  NSFileManager *aFileManager;
  NSString *aPath, *thisMailFile;
  int i, fileCount;
  int numMessages, messageIndex;
  FILE *aStream;
  
  if (theDirectory == nil)
    {
      return NO;
    }
  
  // Get our current count of messages
  numMessages = [[_cacheManager cache] count];
  
  aFileManager = [NSFileManager defaultManager];

  // Read the directory
  aPath = [NSString stringWithFormat: @"%@/%@", _path, theDirectory];
  mailFiles = [[NSMutableArray alloc] initWithArray: [aFileManager directoryContentsAtPath: aPath]];
  AUTORELEASE(mailFiles);

  // We remove Apple Mac OS X .DS_Store file
  [mailFiles removeObject: @".DS_Store"];
  fileCount = [mailFiles count];
  
  if ( mailFiles != nil &&
       fileCount > 0)
    {
      for (i = 0; i < fileCount; i++)
	{
	  thisMailFile = [NSString stringWithFormat: @"%@/%@", aPath, [mailFiles objectAtIndex: i]];
	  aStream = fopen([thisMailFile cString], "r");
	  if ( !aStream )
	    {
	      continue;
	    }
	 
	  messageIndex = [self parse_mbox: thisMailFile  stream: aStream  index: numMessages];
	  
	  if (messageIndex >= 0)
	    {
	      numMessages++;
	    }
	  fclose(aStream);
	  
	  if (messageIndex < 0)
	    {
	      continue;
	    }
	  
	  // If we read this from the "new" or "tmp" sub-directories, move it to the "cur" directory
	  if ([theDirectory isEqualToString: @"new"] || [theDirectory isEqualToString: @"tmp"])
	    {
	      NSString *newPath;
	      LocalMessage *aLocalMessage;
	      
	      newPath = [NSString stringWithFormat: @"%@/cur/%@", _path, [mailFiles objectAtIndex: i]];
	      
	      if ([aFileManager movePath: thisMailFile  toPath: newPath  handler: nil])
		{
		  aLocalMessage = [[_cacheManager cache] objectAtIndex: messageIndex];
		 
		  if (aLocalMessage)
		    {
		      [aLocalMessage setMailFilename: newPath];
		    }
		}
	    }
	  
	}
    }
  
  return YES;
}

@end
