/*
**  LocalFolder+mbox.m
**
**  Copyright (c) 2004
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**  
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**  
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <Pantomime/LocalFolder+mbox.h>

#include <Pantomime/Flags.h>
#include <Pantomime/LocalCacheManager.h>
#include <Pantomime/LocalMessage.h>
#include <Pantomime/LocalStore.h>
#include <Pantomime/NSData+Extensions.h>
#include <Pantomime/NSString+Extensions.h>
#include <Pantomime/Parser.h>

#include <Foundation/NSFileManager.h>
#include <Foundation/NSNotification.h>

#include <sys/file.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>

#warning FIXME
#undef PERFORM_SELECTOR_2

#define PERFORM_SELECTOR_2(sel, name, obj, key) \
if ([(LocalStore *)[self store] delegate] && [[(LocalStore *)[self store] delegate] respondsToSelector: sel]) \
{ \
  [[(LocalStore *)[self store] delegate] performSelector: sel \
	     withObject: [NSNotification notificationWithName: name \
					 object: [self store] \
					 userInfo: [NSDictionary dictionaryWithObject: obj \
								 forKey: key]]]; \
}

//
//
//
@implementation LocalFolder (mbox)

- (void) close_mbox
{
  //NSLog(@"LocalFolder+mbox: -close_mbox");
  fclose(stream);
  flock(fd, LOCK_UN);
  close(fd);	
}


//
//
//
- (void) expunge_mbox: (BOOL) returnDeletedMessages
{
  NSMutableArray *aMutableArray;
  
  FILE *theInputStream, *theOutputStream;

  LocalMessage *aMessage;
  Flags *theFlags;
  
  BOOL writeWasSuccessful, seenStatus, seenXStatus, doneWritingHeaders;
  NSString *pathToMailbox;
  
  int i, messageNumber, count;
  char aLine[1024];
  
  pathToMailbox = [NSString stringWithFormat: @"%@/%@", [_store path], _name];
  
  // The stream is used to store (temporarily) the new local folder
  theOutputStream = fopen([[NSString stringWithFormat: @"%@.tmp", pathToMailbox] cString], "a");
  theInputStream = [self stream];

  // We assume that our write operation was successful and we initialize our messageNumber to 1
  writeWasSuccessful = YES;
  messageNumber = 1;
  
  // We verify it the creation failed
  if (!theOutputStream)
    {
      PERFORM_SELECTOR_2(@selector(folderExpungeFailed:), @"PantomimeFolderExpungeFailed", self, @"Folder");
      return;
    }
  
  aMutableArray = [NSMutableArray array];
  count = [allMessages count];

  for (i = 0; i < count; i++)
    {
      aMessage = [allMessages objectAtIndex: i];
      theFlags = [aMessage flags];

      doneWritingHeaders = seenStatus = seenXStatus = NO;    
      
      if ( [theFlags contain: PantomimeDeleted] )
	{
	  // We add our message to our array of deleted messages, if we need to.
	  if ( returnDeletedMessages )
	    {
	      [aMutableArray addObject: [aMessage rawSource]];
	    }
	  
	  [(LocalCacheManager *)_cacheManager removeObject: aMessage];
	}
      else
	{
	  long delta, position, size;
	  int headers_length;
	  
	  // We get our position and headers_length
	  position = ftell(theOutputStream);
	  headers_length = 0;
	  
	  // We seek to the beginning of the message
	  fseek(theInputStream, [aMessage filePosition], SEEK_SET);
	 
	  size = [aMessage size];
	  memset(aLine, 0, 1024);
	  
	  while ( fgets(aLine, 1024, theInputStream) != NULL &&
		  (ftell(theInputStream) < ([aMessage filePosition] + size)) )
	    {
	      // We verify if we aren't finished reading our headers
	      if ( !doneWritingHeaders )
		{
		  // We check for the "null line" (ie., end of headers)
		  if ( strlen(aLine) == 1 && strcmp("\n", aLine) == 0 )
		    {
		      doneWritingHeaders = YES;
		      
		      if ( !seenStatus ) 
			{
			  fputs( [[NSString stringWithFormat: @"Status: %@\n",
					    [theFlags statusString]] cString], theOutputStream );
			}
		      
		      if ( !seenXStatus ) 
			{
			  fputs( [[NSString stringWithFormat: @"X-Status: %@\n",
						[theFlags xstatusString]] cString], theOutputStream );
			}

		      // Since we are done writing our headers, we update the headers_length variable
		      headers_length = ftell(theOutputStream) - position;

		      // We adjust the size of the message since headers might have been rewritten (Status/X-Status).
		      // We need the trailing -1 to actually remove the header/content separator (a single \n).
		      delta = headers_length - ([aMessage bodyFilePosition] - [aMessage filePosition] - 1);
		      
		      if ( delta > 0 )
			{
			  [aMessage setSize: (size+delta)];
			}
		    }
		  
		  // If we read the Status header, we replace it with the current Status header
		  if (strncasecmp(aLine,"Status:", 7) == 0) 
		    {
		      seenStatus = YES;
		      memset(aLine, 0, 1024);
		      sprintf(aLine, "Status: %s\n", [[theFlags statusString] cString]); 
		    }
		  else if (strncasecmp(aLine,"X-Status:", 9) == 0)
		    {
		      seenXStatus = YES;
		      memset(aLine, 0, 1024);
		      sprintf(aLine, "X-Status: %s\n", [[theFlags xstatusString] cString]); 
		    }
		}
	      
	      // We write our line to our new stream
	      if ( fputs(aLine, theOutputStream) < 0 )
		{
		  writeWasSuccessful = NO;
		  break;
		}
	      
	      memset(aLine, 0, 1024);
	    } // while (...)
	  
	  // We add our message separator
	  if ( fputs("\n", theOutputStream) < 0 )
	    {
	      writeWasSuccessful = NO;
	      break;
	    }
	    
	  // We update our message's ivars
	  [aMessage setFilePosition: position];
	  [aMessage setBodyFilePosition: (position+headers_length+1)];
	  [aMessage setMessageNumber: messageNumber];
	  
	  // We increment our messageNumber local variable
	  messageNumber++;
	}

    } // for (i = 0; i < count; i++)
  
  // We close our output stream
  if ( fclose(theOutputStream) != 0 )
    {
      writeWasSuccessful = NO;
    }
  
  //
  // We verify if the last write was successful, if yes, we remove our original mailbox
  // and we replace it by our temporary mailbox.
  //
  if (writeWasSuccessful)
    {
      // We close the current folder
      fclose(theInputStream);
      flock([self fd], LOCK_UN);
      close([self fd]);
      
      // Now that everything is alright, replace <folder name> by <folder name>.tmp
      [[NSFileManager defaultManager] removeFileAtPath: pathToMailbox
				      handler: nil];
      [[NSFileManager defaultManager] movePath: [NSString stringWithFormat: @"%@.tmp", pathToMailbox]
				      toPath: pathToMailbox
				      handler: nil];
      
      // We sync our cache
      [_cacheManager synchronize];
      
      // Now we re-open our folder and update the 'allMessages' ivar in the Folder superclass
      [self open_mbox];
      [self setMessages: [_cacheManager cache]];
    }
  
  //
  // The last write failed, let's remove our temporary file and keep the original mbox which, might
  // contains non-updated status flags or messages that have been transferred/deleted.
  //
  else
    {
      NSLog(@"Writing to %@ failed. We keep the original mailbox.", pathToMailbox);
      NSLog(@"This can be due to the fact that your partition containing this mailbox is full or that you don't have write permission in the directory where this mailbox is.");
      [[NSFileManager defaultManager] removeFileAtPath: [NSString stringWithFormat: @"%@.tmp", pathToMailbox]
				      handler: nil];
      PERFORM_SELECTOR_2(@selector(folderExpungeFailed:), @"PantomimeFolderExpungeFailed", self, @"Folder");
      return;
    }
  
  PERFORM_SELECTOR_2(@selector(folderExpungeCompleted:), @"PantomimeFolderExpungeCompleted", self, @"Folder");
}



//
//
//
- (FILE *) open_mbox
{
  FILE *aStream;
  
  if (!_path)
    {
      return NULL;
    }

  fd = open([_path cString], O_RDWR);
  
  if (fd < 0)
    {
      NSLog(@"LocalFolder: Unable to get folder descriptor.");
      return NULL;
    }
    
  if (flock(fd, LOCK_EX|LOCK_NB) < 0) 
    {
      NSLog(@"LocalFolder: Unable to obtain the lock on the folder descriptor.");
      close(fd);
      return NULL;
    }
  else 
    {
      flock(fd, LOCK_UN);
    }
  
  aStream = fdopen(fd, "r+");
  
  [self setStream: aStream];
  
  if (aStream == NULL)
    {
      NSLog(@"LocalFolder: Unable to open the specified mailbox.");
      return NULL;
    }
  
  flock(fd, LOCK_EX|LOCK_NB);
  
  return aStream;
}


//
//
//
- (int) parse_mbox: (NSString *) theFile
	    stream: (FILE*) aStream
	     index: (int) theIndex
{
  LocalMessage *aLocalMessage;
  long begin, end, size;
  char aLine[1024];
  int index;
  BOOL success = NO;
  
  // We initialize our variables
  aLocalMessage = [[LocalMessage alloc] init];
  begin = ftell(aStream);
  end = 0L;
  index = theIndex;
  
  while (fgets(aLine, 1024, aStream) != NULL)
    {
      switch ( tolower(aLine[0]) )
	{	  
	case 'c':
	  if (strncasecmp(aLine, "Cc", 2) == 0)
	    {
	      [Parser parseDestination: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      forType: PantomimeCcRecipient
		      inMessage: aLocalMessage];
	    }

	  break;
	  
	case 'd':
	  if (strncasecmp(aLine, "Date", 4) == 0)
	    {
	      [Parser parseDate: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 'f':
	  if (strncasecmp(aLine, "From ", 5) == 0)
	    {
	      // do nothing, it's our message separator
	    }
	  else if (strncasecmp(aLine, "From", 4) == 0)
	    {
	      [Parser parseFrom: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 'i':
	  if (strncasecmp(aLine, "In-Reply-To", 11) == 0)
	    {
	      [Parser parseInReplyTo: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;

	case 'm':
	  if (strncasecmp(aLine, "Message-ID", 10) == 0)
	    {
	      [Parser parseMessageID: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  else if (strncasecmp(aLine, "MIME-Version", 12) == 0)
	    {
	      [Parser parseMimeVersion: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 'r':
	  if (strncasecmp(aLine, "References", 10) == 0)
	    {
	      [Parser parseReferences: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 's':
	  if (strncasecmp(aLine, "Status", 6) == 0)
	    {
	      [Parser parseStatus: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  else if (strncasecmp(aLine, "Subject", 7) == 0)
	    {
	      [Parser parseSubject: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 't':
	  if (strncasecmp(aLine, "To", 2) == 0)
	    {
	      [Parser parseDestination: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      forType: PantomimeToRecipient
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case 'x':
	  if (strncasecmp(aLine, "X-Status", 8) == 0)
	    {
	      [Parser parseXStatus: [self unfoldLinesStartingWith: aLine  fileStream: aStream]
		      inMessage: aLocalMessage];
	    }
	  break;
	  
	case '\n':
	  
	  [aLocalMessage setFilePosition: begin];
	  [aLocalMessage setBodyFilePosition: ftell(aStream)];
	  
	  // We must set this in case the last message of our mbox is
	  // an "empty message", i.e, a message with all the headers but
	  // with an empty content.
	  end = ftell(aStream);
	  
	  while (fgets(aLine, 1024, aStream) != NULL)
	    {
	      if (strncmp(aLine, "From ", 5) == 0) break;
	      else end = ftell(aStream);
	    }
	  
	  fseek(aStream, end, SEEK_SET);
	  size = end - begin;
	  
	  
	  // We increment our index
	  index = index + 1;
	  
	  // We set the properties of our message object and we add it to our folder.
	  [aLocalMessage setSize: size];
	  [aLocalMessage setMessageNumber: index];
	  [aLocalMessage setFolder: self];
	  [aLocalMessage setType: _type];
	  [aLocalMessage setMailFilename: theFile];
	  [self appendMessage: aLocalMessage];

	  // if we are reading a maildir message, check for flag information in the file name
	  if (_type == PantomimeFormatMaildir)
	    {
	      NSString *uniquePattern, *info;
	      int indexOfPatternSeparator;
	      
	      // name of file will be unique_pattern:info with the status flags in the info field
	      indexOfPatternSeparator = [theFile indexOfCharacter: ':'];
	      
	      if (indexOfPatternSeparator > 1)
		{
		  uniquePattern = [theFile substringToIndex: indexOfPatternSeparator];
		  info = [theFile substringFromIndex: indexOfPatternSeparator];
		}
	      else
		{
		  uniquePattern = theFile;
		  info = @"";
		}
	      
	      // remove all the flags and rebuild
	      [[aLocalMessage flags] removeAll];
	      
	      if ([info indexOfCharacter: 'S'] >= 0)
		{
		  [[aLocalMessage flags] add: PantomimeSeen];
		}
	      
	      if ([info indexOfCharacter: 'R'] >= 0)
		{
		  [[aLocalMessage flags] add: PantomimeAnswered];
		}
	      
	      if ([info indexOfCharacter: 'F'] >= 0)
		{
		  [[aLocalMessage flags] add: PantomimeFlagged];
		}
	      
	      if ([info indexOfCharacter: 'D'] >= 0)
		{
		  [[aLocalMessage flags] add: PantomimeDraft];
		}
	      
	      if ([info indexOfCharacter: 'T'] >= 0)
		{
		  [[aLocalMessage flags] add: PantomimeDeleted];
		}
	    }		
	  
	  // We add to our cache
	  [_cacheManager addObject: aLocalMessage];
	  success = YES;		
	  
	  RELEASE(aLocalMessage);
	  
	  begin = ftell(aStream);
	  
	  // We re-init our message and our mutable string for the next message we're gonna read
	  aLocalMessage = [[LocalMessage alloc] init];
	  break;
	  
	default:
	  break;
	}
    }

  
  // We sync our cache
  [_cacheManager synchronize];
  
  RELEASE(aLocalMessage);

  if (!success)
    NSLog(@"Failed to parse mail file %@!", theFile);
  
  // Return the position of the last message read in the cache list.
  return (index-1);
}


//
// This method is used to unfold the lines that have been folded
// by starting with the first line.
//
- (NSData *) unfoldLinesStartingWith: (char *) firstLine 
			  fileStream: (FILE*) theStream
{
  NSMutableData *aMutableData;
  NSData *aData;
  char aLine[1024], buf[1024];
  long mark;
  
  // We initialize our buffers
  memset(aLine, 0, 1024);
  memset(buf, 0, 1024);
    
  mark = ftell(theStream);
  fgets(aLine, 1024, theStream);

  if (aLine == NULL)
    {
      return [NSData dataWithBytes: firstLine  length: strlen(firstLine)];
    }

  // We create our mutable data
  aMutableData = [[NSMutableData alloc] initWithCapacity: strlen(firstLine)];

  // We remove the trailing \n and we append our first line to our mutable data
  strncpy(buf, firstLine, strlen(firstLine) - 1);
  [aMutableData appendCFormat: @"%s ", buf];
  
  // We loop as long as we have a space or tab character as the first character 
  // of the line that we just read
  while ( aLine[0] == 9 || aLine[0] == 32 )
    {
      char *ptr;

      // We skip the first char
      ptr = aLine;
      ptr++;
      
      // We init our buffer and we copy the data into it by trimming the trailing \n
      memset(buf, 0, 1024);
      strncpy(buf, ptr, strlen(ptr) - 1);
      [aMutableData appendCFormat: @"%s ", buf];

      // We set our mark and get the next folded line (if there's one)
      mark = ftell(theStream);
      memset(aLine, 0, 1024);
      fgets(aLine, 1024, theStream);
      
      if (aLine == NULL)
        {
	  RELEASE(aMutableData);
          return nil;
        }
    }

  // We reset our file pointer position and we free our C buffers.
  fseek(theStream, mark, SEEK_SET);
  
  // We trim our last " " that we added to our data
  aData = [aMutableData subdataToIndex: [aMutableData length] - 1];
  
  RELEASE(aMutableData);

  return aData;
}

@end
